""" Top-level imports for the library. """

import sys

from .base import BaseGame, BaseMatching, BasePlayer
from .matchings import MultipleMatching, SingleMatching
from .players import Player
from .version import __version__

if not sys.warnoptions:
    import warnings

    warnings.simplefilter("always")

__all__ = [
    "BaseGame",
    "BaseMatching",
    "BasePlayer",
    "SingleMatching",
    "MultipleMatching",
    "Player",
    "__version__",
]



""" Functions for the SM algorithms. """

from .util import _delete_pair, _match_pair


def _unmatch_pair(suitor, reviewer):
    """ Unmatch a (suitor, reviewer) pair. """

    suitor._unmatch()
    reviewer._unmatch()


def stable_pairing(suitors, reviewers, optimal="suitor"):
    """An extended version of the original Gale-Shapley algorithm which makes
    use of the inherent structures of SM instances. A unique, stable and optimal
    matching is found for any valid set of suitors and reviewers. The optimality
    of the matching is with respect to one party and is subsequently the worst
    stable matching for the other.

    Parameters
    ----------
    suitors : list of Player
        The suitors in the game. Each must rank all of those in ``reviewers``.
    reviewers : list of Player
        The reviewers in the game. Each must rank all of those in ``suitors``.
    optimal : str, optional
        Which party the matching should be optimised for. Must be one of
        ``"suitor"`` and ``"reviewer"``. Defaults to the former.

    Returns
    -------
    matching : Matching
        A dictionary-like object where the keys are given by the members of
        ``suitors``, and the values are their match in ``reviewers``.
    """

    if optimal.lower() == "reviewer":
        suitors, reviewers = reviewers, suitors

    free_suitors = suitors[:]
    while free_suitors:

        suitor = free_suitors.pop()
        reviewer = suitor.get_favourite()

        if reviewer.matching:
            current_match = reviewer.matching
            _unmatch_pair(current_match, reviewer)
            free_suitors.append(current_match)

        _match_pair(suitor, reviewer)

        successors = reviewer.get_successors()
        for successor in successors:
            _delete_pair(successor, reviewer)

    if optimal.lower() == "reviewer":
        suitors, reviewers = reviewers, suitors

    return {s: s.matching for s in suitors}



""" The container class for instances of SA. """

from .base_container import BaseContainer


class Container(BaseContainer):
    """A class to represent a container in an instance of SA.

    Parameters
    ----------
    name : object
        An identifier. This should be unique and descriptive.
    capacity : int
        The maximum number of matches the container can have.

    Attributes
    ----------
    resources : list of Resource
        The resources that the container runs. Defaults to an empty list.
    prefs : list of Player
        The container's preferences. Defaults to ``None`` and is updated via
        the ``set_prefs`` method.
    pref_names : list
        A list of the names in ``prefs``. Updates with ``prefs`` via
        ``set_prefs``.
    matching : list of Player
        The current matches of the container. An empty list if currently
        unsubscribed, and updated through its resources' matching updates.
    """

    def __init__(self, name, capacity):

        super().__init__(name, capacity)
        self.resources = []

    def _forget(self, player):
        """Only forget ``player`` if it is not ranked by any of the
        container's resources."""

        if player in self.prefs and not any(
            [player in resource.prefs for resource in self.resources]
        ):
            prefs = self.prefs[:]
            prefs.remove(player)
            self.prefs = prefs

    def set_prefs(self, players):
        """Set the preference of the container, and pass those on to its
        resources."""

        self.prefs = players
        self._pref_names = [player.name for player in players]
        self._original_prefs = players[:]

        for resource in self.resources:
            acceptable = [
                player for player in players if resource in player.prefs
            ]
            resource.set_prefs(acceptable)

    def get_favourite(self):
        """Find the container's favourite player that it is not currently
        matched to, but has a preference of, one of the container's
        under-subscribed resources. Also return the player's favourite
        under-subscribed resource. If no such player exists, return ``None``.
        """

        if len(self.matching) < self.capacity:
            for player in self.prefs:
                for resource in player.prefs:
                    if (
                        resource.container == self
                        and player not in resource.matching
                        and len(resource.matching) < resource.capacity
                    ):
                        return player, resource

        return None



""" The Resource class for use in instances of SA. """

from .base_container import BaseContainer


class Resource(BaseContainer):
    """A class to represent a resource in an instance of SA.

    Parameters
    ----------
    name : object
        An identifier. This should be unique and descriptive.
    capacity : int
        The maximum number of matches the resource can have.

    Attributes
    ----------
    container : Container
        The container that runs the resource. Defaults to ``None``.
    prefs : list of Player
        The resource's preferences. Inherited from ``container``.
    pref_names : list
        A list of the names in ``prefs``. Updates with ``prefs`` via the
        ``container.set_prefs`` method.
    matching : list of Player
        The current matches of the resource. An empty list if currently
        unsubscribed.
    """

    def __init__(self, name, capacity):

        super().__init__(name, capacity)
        self.container = None

    def _forget(self, player):
        """Remove ``player`` from the preference list of the resource and its
        container."""

        if player in self.prefs:
            prefs = self.prefs[:]
            prefs.remove(player)
            self.prefs = prefs
            self.container._forget(player)

    def _match(self, player):
        """Match the resource to ``player``, and update the resource
        container's matching to include ``player``, too."""

        self.matching.append(player)
        self.matching.sort(key=self.prefs.index)
        self.container._match(player)

    def _unmatch(self, player):
        """Break the matching between the resource and ``player``, and the
        matching between ``player`` and the resource container."""

        matching = self.matching[:]
        matching.remove(player)
        self.matching = matching
        self.container._unmatch(player)

    def set_container(self, container):
        """Set the resource's container and add the resource to their list
        of active resources."""

        self.container = container
        if self not in container.resources:
            container.resources.append(self)



""" The base Player class for use in various games. """

from galeshapley import BasePlayer


class Player(BasePlayer):
    """A class to represent a player within the matching game.

    Parameters
    ----------
    name : object
        An identifier. This should be unique and descriptive.

    Attributes
    ----------
    prefs : list of Player
        The player's preferences. Defaults to ``None`` and is updated using the
        ``set_prefs`` method.
    pref_names : list
        A list of the names in ``prefs``. Updates with ``prefs`` via
        ``set_prefs`` method.
    matching : Player or None
        The current match of the player. ``None`` if not currently matched.
    _original_prefs : list of Player
        The original set of player preferences.
    """

    def _match(self, other):
        """ Assign the player to be matched to some other player. """

        self.matching = other

    def _unmatch(self):
        """ Set the player to be unmatched. """

        self.matching = None

    def get_favourite(self):
        """ Get the player's favourite player. """

        return self.prefs[0]

    def get_successors(self):
        """ Get all the successors to the current match of the player. """

        idx = self.prefs.index(self.matching)
        return self.prefs[idx + 1 :]

    def check_if_match_is_unacceptable(self, unmatched_okay=False):
        """Check the acceptability of the current match, with the stipulation
        that being unmatched is okay (or not)."""

        other = self.matching

        if other is None and unmatched_okay is False:
            return self.unmatched_message()

        elif other is not None and other not in self.prefs:
            return self.not_in_preferences_message(other)



""" Exceptions for game solver checks. """


class MatchingError(Exception):
    """ A generic error for when something is wrong with an internal object. """

    def __init__(self, **kwargs):

        for key, val in kwargs.items():
            self.__setattr__(key, val)

        self.message = kwargs
        super().__init__(self.message)


class NoStableMatchingWarning(UserWarning):
    """ A warning for when a game does not have a complete stable galeshapley. """


class PreferencesChangedWarning(UserWarning):
    """ A warning for when the preferences of a player are invalid. """


class CapacityChangedWarning(UserWarning):
    """ A warning for when the capacity of a player is invalid. """


class PlayerExcludedWarning(UserWarning):
    """ A warning for when a player is to be excluded from a game. """



""" Top-level imports for the `galeshapley.games` subpackage. """

from .base_container_part import BaseContainerPart
from .stable_pairing import StablePairing
from .stable_duos import StableDuos
from .player_allocation import PlayerAllocation

__all__ = [
    "BaseContainerPart",
    "StablePairing",
    "StableDuos",
    "PlayerAllocation",
]



""" Functions for the HR algorithms. """

from .util import _delete_pair, _match_pair


def _unmatch_pair(part, base_container):
    """ Unmatch a (part, base_container)-pair. """

    part._unmatch()
    base_container._unmatch(part)


def _check_available(base_container):
    """ Check whether a base_container is willing and able to take an applicant. """

    return len(base_container.matching) < base_container.capacity and set(
        base_container.prefs
    ).difference(base_container.matching)


def base_container_part(parts, base_containers, optimal="part"):
    """Solve an instance of HR using an adapted Gale-Shapley algorithm
    :cite:`Rot84`. A unique, stable and optimal matching is found for the given
    set of parts and base_containers. The optimality of the matching is found with
    respect to one party and is subsequently the worst stable matching for the
    other.

    Parameters
    ----------
    parts : list of Player
        The parts in the game. Each part must rank a non-empty subset
        of the elements of ``base_containers``.
    base_containers : list of BaseContainer
        The base_containers in the game. Each base_container must rank all the parts
        that have ranked them.
    optimal : str, optional
        Which party the matching should be optimised for. Must be one of
        ``"part"`` and ``"base_container"``. Defaults to the former.

    Returns
    -------
    matching : Matching
        A dictionary-like object where the keys are the members of
        ``base_containers``, and the values are their matches ranked by preference.
    """

    if optimal == "part":
        return part_optimal(parts, base_containers)
    if optimal == "base_container":
        return base_container_optimal(base_containers)


def part_optimal(parts, base_containers):
    """Solve the instance of HR to be part-optimal. The algorithm is as
    follows:

        0. Set all parts to be unmatched, and all base_containers to be totally
        unsubscribed.

        1. Take any unmatched part with a non-empty preference list,
        :math:`r`, and consider their most preferred base_container, :math:`h`. Match
        them to one another.

        2. If, as a result of this new matching, :math:`h` is now
        over-subscribed, find the worst part currently assigned to
        :math:`h`, :math:`r'`. Set :math:`r'` to be unmatched and remove them
        from :math:`h`'s galeshapley. Otherwise, go to 3.

        3. If :math:`h` is at capacity (fully subscribed) then find their worst
        current match :math:`r'`. Then, for each successor, :math:`s`, to
        :math:`r'` in the preference list of :math:`h`, delete the pair
        :math:`(s, h)` from the game. Otherwise, go to 4.

        4. Go to 1 until there are no such parts left, then end.
    """

    free_parts = parts[:]
    while free_parts:

        part = free_parts.pop()
        base_container = part.get_favourite()

        if len(base_container.matching) == base_container.capacity:
            worst = base_container.get_worst_match()
            _unmatch_pair(worst, base_container)
            free_parts.append(worst)

        _match_pair(part, base_container)

        if len(base_container.matching) == base_container.capacity:
            successors = base_container.get_successors()
            for successor in successors:
                _delete_pair(base_container, successor)
                if not successor.prefs:
                    free_parts.remove(successor)

    return {r: r.matching for r in base_containers}


def base_container_optimal(base_containers):
    """Solve the instance of HR to be base_container-optimal. The algorithm is as
    follows:

        0. Set all parts to be unmatched, and all base_containers to be totally
        unsubscribed.

        1. Take any base_container, :math:`h`, that is under-subscribed and whose
        preference list contains any part they are not currently assigned
        to, and consider their most preferred such part, :math:`r`.

        2. If :math:`r` is currently matched, say to :math:`h'`, then unmatch
        them from one another. In any case, match :math:`r` to :math:`h` and go
        to 3.

        3. For each successor, :math:`s`, to :math:`h` in the preference list of
        :math:`r`, delete the pair :math:`(r, s)` from the game.

        4. Go to 1 until there are no such base_containers left, then end.
    """

    free_base_containers = base_containers[:]
    while free_base_containers:

        base_container = free_base_containers.pop()
        part = base_container.get_favourite()

        if part.matching:
            current_match = part.matching
            _unmatch_pair(part, current_match)
            if current_match not in free_base_containers:
                free_base_containers.append(current_match)

        _match_pair(part, base_container)
        if _check_available(base_container):
            free_base_containers.append(base_container)

        successors = part.get_successors()
        for successor in successors:
            _delete_pair(part, successor)
            if not _check_available(successor) and successor in free_base_containers:
                free_base_containers.remove(successor)

    return {r: r.matching for r in base_containers}



""" Top-level imports for the `galeshapley.algorithms` subpackage. """

from .base_container_part import base_container_part
from .stable_pairing import stable_pairing
from .stable_duos import stable_duos
from .player_allocation import player_allocation

__all__ = [
    "base_container_part",
    "stable_pairing",
    "stable_duos",
    "player_allocation",
]



""" Useful functions for the running of the various core algorithms. """


def _delete_pair(player, other):
    """Make a player forget another (and vice versa), deleting the pair from
    further consideration in the game."""

    player._forget(other)
    other._forget(player)


def _match_pair(player, other):
    """Match the players given by `player` and `other`."""

    player._match(other)
    other._match(player)



""" The BaseContainer class for use in instances of HR. """

from galeshapley import BasePlayer


class BaseContainer(BasePlayer):
    """A class to represent a base_container in an instance of HR. Also used as a
    parent class to ``Resource`` and ``Container``.

    Parameters
    ----------
    name : object
        An identifier. This should be unique and descriptive.
    capacity : int
        The maximum number of matches the base_container can have.
    _original_capacity : int
        A record of the player's original capacity in case it is altered when
        passed to a game.

    Attributes
    ----------
    prefs : list of Player
        The base_container's preferences. Defaults to ``None`` and is updated using
        the ``set_prefs`` method.
    _original_prefs : list of Player
        A record of the player's original preferences.
    pref_names : list
        A list of the names in ``prefs``. Updates with ``prefs`` via the
        ``set_prefs`` method.
    matching : list of Player
        The current matches of the base_container. An empty list if currently
        unsubscribed.
    """

    def __init__(self, name, capacity):

        super().__init__(name)
        self._capacity = capacity
        self._original_capacity = capacity
        self.matching = []

    def _match(self, part):
        """ Add ``part`` to the base_container's matching, and then sort it. """

        self.matching.append(part)
        self.matching.sort(key=self.prefs.index)

    def _unmatch(self, part):
        """ Remove ``part`` from the base_container's galeshapley. """

        matching = self.matching[:]
        matching.remove(part)
        self.matching = matching

    def oversubscribed_message(self):

        return (
            f"{self} is matched to {self.matching} which is over their "
            f"capacity of {self.capacity}."
        )

    def get_favourite(self):
        """Get the base_container's favourite part with whom they are not
        currently matched. If no such part exists, return ``None``."""

        for player in self.prefs:
            if player not in self.matching:
                return player

        return None

    def get_worst_match(self):
        """Get the player's worst current match. This assumes that the matching
        is in order of preference."""

        return self.matching[-1]

    def get_successors(self):
        """ Get the successors to the player's worst current match. """

        idx = self.prefs.index(self.get_worst_match())
        return self.prefs[idx + 1 :]

    def get_capacity(self):
        if callable(self._capacity):
            return self._capacity()
        else:
            return self._capacity

    def set_capacity(self, c):
        self._capacity = c

    def del_capacity(self):
        del self._capacity

    def check_if_match_is_unacceptable(self, **kwargs):
        """ Check the acceptability of the current matches. """

        issues = []
        for other in self.matching:
            if other not in self.prefs:
                issues.append(self.not_in_preferences_message(other))

        return issues

    def check_if_oversubscribed(self):
        """ Check whether the player has too many matches. """

        if len(self.matching) > self.capacity:
            return self.oversubscribed_message()

        return False

    capacity = property(get_capacity, set_capacity, del_capacity)


""" The SM game class and supporting functions. """
import copy

from galeshapley import BaseGame, Player, SingleMatching
from galeshapley.algorithms import stable_pairing
from galeshapley.exceptions import MatchingError


class StablePairing(BaseGame):
    """A class for solving instances of the stable marriage problem (SM).

    Parameters
    ----------
    suitors : list of Player
        The suitors in the game. Each suitor must rank all elements in
        ``reviewers``.
    reviewers : list of Player
        The reviewers in the game. Each reviewer must rank all elements in
        ``suitors``.

    Attributes
    ----------
    matching : Matching or None
        Once the game is solved, a matching is available. This uses the suitors
        and reviewers as keys and values, respectively, in a ``Matching``
        object. Initialises as `None`.
    blocking_pairs : list of (Player, Player)
        The suitor-reviewer pairs that both prefer one another to their current
        match. Initialises as ``None``.
    """

    def __init__(self, suitors, reviewers):

        suitors, reviewers = copy.deepcopy([suitors, reviewers])
        self.suitors = suitors
        self.reviewers = reviewers

        super().__init__()
        self.check_inputs()

    @classmethod
    def create_from_dictionaries(cls, suitor_prefs, reviewer_prefs):
        """ Create an instance of SM from two preference dictionaries. """

        suitors, reviewers = _make_players(suitor_prefs, reviewer_prefs)
        game = cls(suitors, reviewers)

        return game

    def solve(self, optimal="suitor"):
        """Solve the instance of SM using either the suitor- or
        reviewer-oriented Gale-Shapley algorithm. Return the galeshapley."""

        self.matching = SingleMatching(
            stable_pairing(self.suitors, self.reviewers, optimal)
        )
        return self.matching

    def check_validity(self):
        """ Check whether the current matching is valid. """

        unmatched_issues = self._check_for_unmatched_players()
        not_in_matching_issues = self._check_for_players_not_in_matching()
        inconsistency_issues = self._check_for_inconsistent_matches()

        if unmatched_issues or not_in_matching_issues or inconsistency_issues:
            raise MatchingError(
                unmatched_players=unmatched_issues,
                players_not_in_matching=not_in_matching_issues,
                inconsistent_matches=inconsistency_issues,
            )

        return True

    def check_stability(self):
        """Check for the existence of any blocking pairs in the current
        matching, thus determining the stability of the galeshapley."""

        blocking_pairs = []
        for suitor in self.suitors:
            for reviewer in self.reviewers:
                if suitor.prefers(
                    reviewer, suitor.matching
                ) and reviewer.prefers(suitor, reviewer.matching):
                    blocking_pairs.append((suitor, reviewer))

        self.blocking_pairs = blocking_pairs
        return not any(blocking_pairs)

    def _check_for_unmatched_players(self):
        """ Check everyone has a match. """

        issues = []
        for player in self.suitors + self.reviewers:
            issue = player.check_if_match_is_unacceptable(unmatched_okay=False)
            if issue:
                issues.append(issue)

        return issues

    def _check_for_players_not_in_matching(self):
        """ Check that everyone appears in the galeshapley. """

        players_in_matching = set(self.matching.keys()) | set(
            self.matching.values()
        )

        issues = []
        for player in self.suitors + self.reviewers:
            if player not in players_in_matching:
                issues.append(f"{player} does not appear in galeshapley.")

        return issues

    def _check_for_inconsistent_matches(self):
        """Check that the game matching is consistent with those of the
        players."""

        issues = []
        for suitor, reviewer in self.matching.items():
            if suitor.matching != reviewer:
                issues.append(
                    f"{suitor} is matched to {suitor.matching} but the "
                    f"matching says they should be matched to {reviewer}."
                )

        return issues

    def check_inputs(self):
        """Raise an error if any of the conditions of the game have been
        broken."""

        self._check_num_players()
        for suitor in self.suitors:
            self._check_player_ranks(suitor)
        for reviewer in self.reviewers:
            self._check_player_ranks(reviewer)

    def _check_num_players(self):
        """ Check that the number of suitors and reviewers are equal. """

        if len(self.suitors) != len(self.reviewers):
            raise ValueError(
                "There must be an equal number of suitors and reviewers."
            )

        return True

    def _check_player_ranks(self, player):
        """ Check that a player has ranked all of the other group. """

        others = self.reviewers if player in self.suitors else self.suitors
        if set(player.prefs) != set(others):
            raise ValueError(
                "Every player must rank each name from the other group. "
                f"{player}: {player.prefs} != {others}"
            )

        return True


def _make_players(suitor_prefs, reviewer_prefs):
    """Make a set of ``Player`` instances each for suitors and reviewers from
    the dictionaries given. Add their preferences."""

    suitor_dict, reviewer_dict = _make_instances(suitor_prefs, reviewer_prefs)

    for suitor_name, suitor in suitor_dict.items():
        prefs = [reviewer_dict[name] for name in suitor_prefs[suitor_name]]
        suitor.set_prefs(prefs)

    for reviewer_name, reviewer in reviewer_dict.items():
        prefs = [suitor_dict[name] for name in reviewer_prefs[reviewer_name]]
        reviewer.set_prefs(prefs)

    suitors = list(suitor_dict.values())
    reviewers = list(reviewer_dict.values())

    return suitors, reviewers


def _make_instances(suitor_prefs, reviewer_prefs):
    """ Create ``Player`` instances for the names in each dictionary. """

    suitor_dict, reviewer_dict = {}, {}
    for suitor_name in suitor_prefs:
        suitor = Player(name=suitor_name)
        suitor_dict[suitor_name] = suitor
    for reviewer_name in reviewer_prefs:
        reviewer = Player(name=reviewer_name)
        reviewer_dict[reviewer_name] = reviewer

    return suitor_dict, reviewer_dict



""" A collection of dictionary-like objects for storing matchings. """
from galeshapley import BaseMatching
from galeshapley.players import Player


class SingleMatching(BaseMatching):
    """A dictionary-like object for storing and updating a matching with
    singular matches such as those in an instance of SM or SR.

    Parameters
    ----------
    dictionary
        The dictionary of matches. Made up of :code:`Player, Optional[Player]`
        key, value pairs.
    """

    def __init__(self, dictionary):

        super().__init__(dictionary)

    def __setitem__(self, player, new):

        self._check_player_in_keys(player)
        self._check_new_valid_type(new, (type(None), Player))

        player.matching = new
        if isinstance(new, Player):
            new.matching = player

        self._data[player] = new


class MultipleMatching(BaseMatching):
    """A dictionary-like object for storing and updating a matching with
    multiple matches such as those in an instance of HR or SA.

    Parameters
    ----------
    dictionary
        The dictionary of matches. Made up of :code:`BaseContainer, List[Player]`
        key, value pairs.
    """

    def __init__(self, dictionary):

        super().__init__(dictionary)

    def __setitem__(self, player, new):

        self._check_player_in_keys(player)
        self._check_new_valid_type(new, (list, tuple))
        for other in new:
            self._check_new_valid_type(other, Player)

        player.matching = new
        for other in new:
            other.matching = player

        self._data[player] = new



""" Functions for the SA algorithm. """

from .util import _delete_pair, _match_pair


def unmatch_pair(player, resource):
    """ Unmatch a player-resource pair. """

    player._unmatch()
    resource._unmatch(player)


def player_allocation(players, resources, containers, optimal="player"):
    """Solve an instance of SA by treating it as a bi-level HR. A unique,
    stable and optimal matching is found for the given set of players, resources
    and containers. The optimality of the matching is found with respect to one
    party and is subsequently the worst stable matching for the other.

    Parameters
    ----------
    players : list of Player
        The players in the game. Each player must rank a subset of the
        elements of ``resources``.
    resources : list of Resource
        The resources in the game. Each resource is offered by a container that
        governs its preferences.
    container : list of Container
        The containers in the game. Each container offers a unique subset of
        ``resources`` and ranks all the players that have ranked at least one of
        these resources.
    optimal : str, optional
        Which party the matching should be optimised for. Must be one of
        ``"player"`` and ``"container"``. Defaults to the former.

    Returns
    =======
    matching : Matching
        A dictionary-like object where the keys are the members of ``resources``
        and their player matches are the values.
    """

    if optimal == "player":
        return player_optimal(players, resources)
    if optimal == "container":
        return container_optimal(resources, containers)


def player_optimal(players, resources):
    """Solve the instance of SA to be player-optimal. The algorithm is as
    follows:

        0. Set all players to be unassigned, and every resource (and container)
        to be totally unsubscribed.

        1. Take any player, :math:`s`, that is unassigned and has a non-empty
        preference list, and consider their most preferred resource, :math:`p`.
        Let :math:`f` denote the container that offers :math:`p`. Assign
        :math:`s` to be matched to :math:`p` (and thus :math:`f`).

        2. If :math:`p` is now over-subscribed, find its worst current match,
        :math:`s'`. Unmatch :math:`p` and :math:`s'`. Else if :math:`f` is
        over-subscribed, find their worst current match, :math:`s''`, and the
        resource they are currently subscribed to, :math:`p'`. Unmatch :math:`p'`
        and :math:`s''`.

        3. If :math:`p` is now at capacity, find their worst current match,
        :math:`s'`. For each successor, :math:`t`, to :math:`s'` in the
        preference list of :math:`p`, delete the pair :math:`(p, t)` from the
        game.

        4. If :math:`f` is at capacity, find their worst current match,
        :math:`s'`. For each successor, :math:`t`, to :math:`s'` in the
        preference list of :math:`f`, for each resource, :math:`p'`, offered by
        :math:`f` that :math:`t` finds acceptable, delete the pair
        :math:`(p', t)` from the game.

        5. Go to 1 until there are no such players left, then end.
    """

    free_players = players[:]
    while free_players:

        player = free_players.pop()
        resource = player.get_favourite()
        container = resource.container

        _match_pair(player, resource)

        if len(resource.matching) > resource.capacity:
            worst = resource.get_worst_match()
            unmatch_pair(worst, resource)
            free_players.append(worst)

        elif len(container.matching) > container.capacity:
            worst = container.get_worst_match()
            worst_resource = worst.matching
            unmatch_pair(worst, worst_resource)
            free_players.append(worst)

        if len(resource.matching) == resource.capacity:
            successors = resource.get_successors()
            for successor in successors:
                _delete_pair(resource, successor)
                if not successor.prefs:
                    free_players.remove(successor)

        if len(container.matching) == container.capacity:
            successors = container.get_successors()
            for successor in successors:

                container_resources = [
                    resource
                    for resource in container.resources
                    if resource in successor.prefs
                ]

                for resource in container_resources:
                    _delete_pair(resource, successor)
                if not successor.prefs:
                    free_players.remove(successor)

    return {p: p.matching for p in resources}


def container_optimal(resources, containers):
    """Solve the instance of SA to be container-optimal. The algorithm is as
    follows:

        0. Set all players to be unassigned, and every resource (and container)
        to be totally unsubscribed.

        1. Take any container member, :math:`f`, that is under-subscribed and
        whose preference list contains at least one player that is not
        currently matched to at least one acceptable (though currently
        under-subscribed) resource offered by :math:`f`. Consider the
        container's most preferred such player, :math:`s`, and that player's
        most preferred such resource, :math:`p`.

        2. If :math:`s` is matched to some other resource, :math:`p'`, then
        unmatch them. In any case, match :math:`s` and :math:`p` (and thus
        :math:`f`).

        3. For each successor, :math:`p'`, to :math:`p` in the preference list
        of :math:`s`, delete the pair :math:`(p', s)` from the game.

        4. Go to 1 until there are no such containers, then end.
    """

    free_containers = containers[:]
    while free_containers:

        container = free_containers.pop()
        player, resource = container.get_favourite()

        if player.matching:
            curr_match = player.matching
            unmatch_pair(player, curr_match)

        _match_pair(player, resource)

        successors = player.get_successors()
        for successor in successors:
            _delete_pair(player, successor)

        free_containers = [
            container
            for container in containers
            if container.get_favourite() is not None
        ]

    return {p: p.matching for p in resources}



""" The SA game class and supporting functions. """
import copy
import warnings

from galeshapley import MultipleMatching
from galeshapley import Player as Player
from galeshapley.algorithms import player_allocation
from galeshapley.exceptions import (
    CapacityChangedWarning,
    MatchingError,
    PreferencesChangedWarning,
)
from galeshapley.games import BaseContainerPart
from galeshapley.players import Resource, Container


class PlayerAllocation(BaseContainerPart):
    """A class for solving instances of the player-allocation problem (SA)
    using an adapted Gale-Shapley algorithm.

    In this case, a blocking pair is defined as any player-resource pair that
    satisfies **all** of the following:

    1. The player has a preference of the resource.
    2. Either the player is unmatched, or they prefer the resource to their
       current resource.
    3. At least one of the following:

       - The resource or its container is under-subscribed.
       - The resource is under-subscribed and the container is at capacity, and
         the player is matched to a resource offered by the container or the
         container prefers the player to its worst currently matched player.
       - The resource is at capacity and its container prefers the player to
         its worst currently matched player.

    Parameters
    ----------
    players : list of Player
        The players in the game. Each player must rank a subset of the
        resources.
    resources : list of Resource
        The resources in the game. Each resource has a container associated
        with it that governs its preferences.
    containers : list of Container
        The containers in the game. Each container oversees a unique subset
        of ``resources`` and ranks all of those players that have ranked at
        least one of its resources.
    clean : bool
        An indicator as to whether the players passed to the game should be
        cleaned in a reductive fashion. Defaults to :code:`False`.

    Attributes
    ----------
    matching : Matching or None
        Once the game is solved, a matching is available. This ``Matching``
        object behaves much like a dictionary that uses the elements of
        ``resources`` as keys and their player matches as values. Initialises as
        ``None``.
    blocking_pairs : list of (Player, Resource)
        Initialises as None. Otherwise, a list of the player-resource blocking
        pairs.
    """

    def __init__(self, players, resources, containers, clean=False):
        
        players, resources, containers = copy.deepcopy(
            [players, resources, containers]
        )
        self.players = players
        self.resources = resources
        self.containers = containers

        self._all_players = players
        self._all_resources = resources
        self._all_containers = containers

        self.clean = clean

        super().__init__(players, resources, clean)
        self.check_inputs()

    def _remove_player(self, player, player_party, other_party=None):
        """Remove players from the game normally unless the player is a
        container."""

        if player_party == "containers":
            self.containers.remove(player)
            for resource in player.resources:
                try:
                    super()._remove_player(resource, "resources", "players")
                except ValueError:
                    pass

        else:
            super()._remove_player(player, player_party, other_party)

    @classmethod
    def create_from_dictionaries(
        cls,
        player_prefs,
        container_prefs,
        resource_containers,
        resource_capacities,
        container_capacities,
        clean=False,
    ):
        """Create an instance of SA from two preference dictionaries,
        affiliations and capacities."""

        players, resources, containers = _make_players(
            player_prefs,
            container_prefs,
            resource_containers,
            resource_capacities,
            container_capacities,
        )
        game = cls(players, resources, containers, clean)

        return game

    def solve(self, optimal="player"):
        """Solve the instance of SA using either the player- or
        container-optimal algorithm."""

        self.matching = MultipleMatching(
            player_allocation(
                self.players, self.resources, self.containers, optimal
            )
        )
        return self.matching

    def check_validity(self):
        """Check whether the current matching is valid. Raise a `MatchingError`
        detailing the issues if not."""

        unacceptable_issues = (
            self._check_for_unacceptable_matches("players")
            + self._check_for_unacceptable_matches("resources")
            + self._check_for_unacceptable_matches("containers")
        )

        oversubscribed_issues = self._check_for_oversubscribed_players(
            "resources"
        ) + self._check_for_oversubscribed_players("containers")

        if unacceptable_issues or oversubscribed_issues:
            raise MatchingError(
                unacceptable_matches=unacceptable_issues,
                oversubscribed_players=oversubscribed_issues,
            )

        return True

    def check_stability(self):
        """Check for the existence of any blocking pairs in the current
        matching, thus determining the stability of the galeshapley."""

        blocking_pairs = []
        for player in self.players:
            for resource in self.resources:
                if (
                    resource in player.prefs
                    and _check_player_unhappy(player, resource)
                    and _check_resource_unhappy(resource, player)
                ):
                    blocking_pairs.append((player, resource))

        self.blocking_pairs = blocking_pairs
        return not any(blocking_pairs)

    def check_inputs(self):
        """Give out warnings if any of the conditions of the game have been
        broken. If the :code:`clean` attribute is :code:`True`, then remove any
        such situations from the game."""

        self._check_inputs_player_prefs_unique("players")
        self._check_inputs_player_prefs_unique("resources")
        self._check_inputs_player_prefs_unique("containers")

        self._check_inputs_player_prefs_all_in_party("players", "resources")
        self._check_inputs_player_prefs_nonempty("players", "resources")

        self._check_inputs_player_prefs_all_in_party("containers", "players")
        self._check_inputs_player_prefs_nonempty("containers", "players")

        self._check_inputs_player_prefs_all_reciprocated("resources")
        self._check_inputs_player_reciprocated_all_prefs("resources", "players")
        self._check_inputs_player_prefs_nonempty("resources", "players")

        self._check_inputs_player_prefs_all_reciprocated("containers")
        self._check_inputs_player_reciprocated_all_prefs(
            "containers", "players"
        )
        self._check_inputs_player_prefs_nonempty("containers", "players")

        self._check_inputs_player_capacity("resources", "players")
        self._check_inputs_player_capacity("containers", "players")
        self._check_inputs_container_capacities_sufficient()
        self._check_inputs_container_capacities_necessary()

    def _check_inputs_player_prefs_all_reciprocated(self, party):
        """Check that each player in :code:`party` has ranked only those
        players that have ranked it, directly or via a resource."""

        if party == "containers":
            for container in self.containers:
                for player in container.prefs:
                    player_prefs_containers = {
                        p.container for p in player.prefs
                    }
                    if container not in player_prefs_containers:
                        warnings.warn(
                            PreferencesChangedWarning(
                                f"{container} ranked {player} but they did "
                                "not rank any of their resources."
                            )
                        )

                        if self.clean:
                            for resource in container.resources:
                                resource._forget(player)

        else:
            super()._check_inputs_player_prefs_all_reciprocated(party)

    def _check_inputs_player_reciprocated_all_prefs(self, party, other_party):
        """Check that each player in :code:`party` has ranked all those players
        in :code:`other_party` that ranked it, directly or via a resource."""

        if party == "containers":
            for container in self.containers:

                players_that_ranked = [
                    player
                    for player in self.players
                    if any(
                        resource in player.prefs
                        for resource in container.resources
                    )
                ]

                for player in players_that_ranked:
                    if player not in container.prefs:
                        warnings.warn(
                            PreferencesChangedWarning(
                                f"{player} ranked a resource provided by "
                                f"{container} but they did not."
                            )
                        )

                        if self.clean:
                            for resource in set(container.resources) & set(
                                player.prefs
                            ):
                                player._forget(resource)

        else:
            super()._check_inputs_player_reciprocated_all_prefs(
                party, other_party
            )

    def _check_inputs_container_capacities_sufficient(self):
        """Check that each container has the capacity to support its largest
        resource(s)."""

        for container in self.containers:

            for resource in container.resources:
                if resource.capacity > container.capacity:
                    warnings.warn(
                        CapacityChangedWarning(
                            f"{resource} has a capacity of {resource.capacity} "
                            "but its container has a capacity of "
                            f"{container.capacity}."
                        )
                    )

                    if self.clean:
                        resource.capacity = container.capacity

    def _check_inputs_container_capacities_necessary(self):
        """Check that each container has at most the necessary capacity for
        all of their resources."""

        for container in self.containers:

            total_resource_capacity = sum(
                p.capacity for p in container.resources
            )

            if container.capacity > total_resource_capacity:
                warnings.warn(
                    CapacityChangedWarning(
                        f"{container} has a capacity of {container.capacity} "
                        "but their resources have a capacity of "
                        f"{total_resource_capacity}"
                    )
                )

                if self.clean:
                    container.capacity = total_resource_capacity


def _check_player_unhappy(player, resource):
    """Determine whether ``player`` is unhappy either because they are
    unmatched or because they prefer ``resource`` to their current galeshapley."""

    return player.matching is None or player.prefers(
        resource, player.matching
    )


def _check_resource_unhappy(resource, player):
    """Determine whether ``resource`` is unhappy because either:
    - they and their container are under-subscribed;
    - they are under-subscribed, their container is full, and either
      ``player`` is in the container's matching or the container prefers
      ``player`` to their worst current matching;
    - ``resource`` is full and their container prefers ``player`` to the
      worst player in the matching of ``resource``.
    """

    container = resource.container

    resource_undersubscribed = len(resource.matching) < resource.capacity
    both_undersubscribed = (
        resource_undersubscribed
        and len(container.matching) < container.capacity
    )

    container_full = len(container.matching) == container.capacity

    swap_available = (
        player in container.matching and player.matching != resource
    ) or container.prefers(player, container.get_worst_match())

    resource_upsetting_container = len(
        resource.matching
    ) == resource.capacity and container.prefers(
        player, resource.get_worst_match()
    )

    return (
        both_undersubscribed
        or (resource_undersubscribed and container_full and swap_available)
        or resource_upsetting_container
    )


def _make_players(
    player_prefs,
    container_prefs,
    resource_containers,
    resource_capacities,
    container_capacities,
):
    """Make a set of ``Player``, ``Resource`` and ``Container`` instances,
    respectively for the players, resources and containers from the
    dictionaries given, and add their preferences."""

    player_dict, resource_dict, container_dict = _make_instances(
        player_prefs,
        resource_containers,
        resource_capacities,
        container_capacities,
    )

    for name, player in player_dict.items():
        prefs = [resource_dict[resource] for resource in player_prefs[name]]
        player.set_prefs(prefs)

    for name, container in container_dict.items():
        prefs = [player_dict[player] for player in container_prefs[name]]
        container.set_prefs(prefs)

    players = list(player_dict.values())
    resources = list(resource_dict.values())
    containers = list(container_dict.values())

    return players, resources, containers


def _make_instances(
    player_prefs,
    resource_containers,
    resource_capacities,
    container_capacities,
):
    """Create ``Player``, ``Resource`` and ``Container`` instances for the
    names in each dictionary."""

    player_dict, resource_dict, container_dict = {}, {}, {}

    for player_name in player_prefs:
        player = Player(name=player_name)
        player_dict[player_name] = player

    for resource_name, container_name in resource_containers.items():
        capacity = resource_capacities[resource_name]
        resource = Resource(name=resource_name, capacity=capacity)
        resource_dict[resource_name] = resource

    for container_name, capacity in container_capacities.items():
        container = Container(name=container_name, capacity=capacity)
        container_dict[container_name] = container

    for resource_name, container_name in resource_containers.items():
        resource = resource_dict[resource_name]
        container = container_dict[container_name]
        resource.set_container(container)

    return player_dict, resource_dict, container_dict



""" Top-level imports for the `galeshapley.players` subpackage. """

from .base_container import BaseContainer
from .player import Player
from .resource import Resource
from .container import Container

__all__ = ["BaseContainer", "Player", "Resource", "Container"]



""" The SR game class and supporting functions. """
import copy

from galeshapley import BaseGame, Player, SingleMatching
from galeshapley.algorithms import stable_duos
from galeshapley.exceptions import MatchingError


class StableDuos(BaseGame):
    """A class for solving instances of the stable duos problem (SR).

    Parameters
    ----------
    players : list of Player
        The players in the game. Each must rank all other players.

    Attributes
    ----------
    matching : Matching or None
        Once the game is solved, a matching is available. This uses the players
        as keys and values in a ``Matching`` object. Initialises as ``None``.
    """

    def __init__(self, players):

        players = copy.deepcopy(players)
        self.players = players

        super().__init__()
        self.check_inputs()

    @classmethod
    def create_from_dictionary(cls, player_prefs):
        """ Create an instance of SR from a preference dictionary. """

        players = _make_players(player_prefs)
        game = cls(players)

        return game

    def solve(self):
        """Solve the instance of SR using Irving's algorithm. Return the
        galeshapley."""

        self.matching = SingleMatching(stable_duos(self.players))
        return self.matching

    def check_validity(self):
        """Check whether the current matching is valid. Raise `MatchingError`
        detailing the issues if not."""

        issues = []
        for player in self.players:
            issue = player.check_if_match_is_unacceptable(unmatched_okay=False)
            if issue:
                issues.append(issue)

        if issues:
            raise MatchingError(unmatched_players=issues)

        return True

    def check_stability(self):
        """Check for the existence of any blocking pairs in the current
        galeshapley. Then the stability of the matching holds when there are no
        blocking pairs and all players have been matched."""

        if None in self.matching.values():
            return False

        blocking_pairs = []
        for player in self.players:
            others = [p for p in self.players if p != player]
            for other in others:
                if (other, player) not in blocking_pairs:
                    both_matched = player.matching and other.matching
                    prefer_each_other = player.prefers(
                        other, player.matching
                    ) and other.prefers(player, other.matching)
                    if both_matched and prefer_each_other:
                        blocking_pairs.append((player, other))

        self.blocking_pairs = blocking_pairs
        return not any(blocking_pairs)

    def check_inputs(self):
        """ Check that all players have ranked all other players. """

        for player in self.players:
            others = {p for p in self.players if p != player}
            if set(player.prefs) != others:
                raise ValueError(
                    "Every player must rank all other players. "
                    f"{player}: {player.prefs} is not a permutation of {others}"
                )

        return True


def _make_players(player_prefs):
    """Make a set of ``Player`` instances from the dictionary given. Add their
    preferences."""

    player_dict = {}
    for player_name in player_prefs:
        player = Player(name=player_name)
        player_dict[player_name] = player

    for player_name, player in player_dict.items():
        prefs = [player_dict[name] for name in player_prefs[player_name]]
        player.set_prefs(prefs)

    players = list(player_dict.values())

    return players



""" Abstract base classes for inheritance. """
import abc
import warnings

from galeshapley.exceptions import PlayerExcludedWarning, PreferencesChangedWarning


class BasePlayer:
    """An abstract base class to represent a player within a matching game.

    Parameters
    ----------
    name : object
        An identifier. This should be unique and descriptive.

    Attributes
    ----------
    prefs : List[BasePlayer]
        The player's preferences. Defaults to ``None`` and is updated using the
        ``set_prefs`` method.
    matching : Optional[BasePlayer]
        The current match of the player. ``None`` if not currently matched.
    _pref_names : Optional[List]
        A list of the names in ``prefs``. Updates with ``prefs`` via
        ``set_prefs`` method.
    _original_prefs : Optional[List[BasePlayer]]
        The original set of player preferences. Defaults to ``None`` and does
        not update after the first ``set_prefs`` method call.
    """

    def __init__(self, name):

        self.name = name
        self.prefs = []
        self.matching = None

        self._pref_names = []
        self._original_prefs = None

    def __repr__(self):

        return str(self.name)

    def _forget(self, other):
        """Forget another player by removing them from the player's preference
        list."""

        prefs = self.prefs[:]
        prefs.remove(other)
        self.prefs = prefs

    def unmatched_message(self):

        return f"{self} is unmatched."

    def not_in_preferences_message(self, other):

        return (
            f"{self} is matched to {other} but they do not appear in their "
            f"preference list: {self.prefs}."
        )

    def set_prefs(self, players):
        """ Set the player's preferences to be a list of players. """

        self.prefs = players
        self._pref_names = [player.name for player in players]

        if self._original_prefs is None:
            self._original_prefs = players[:]

    def prefers(self, player, other):
        """Determines whether the player prefers a player over some other
        player."""

        prefs = self._original_prefs
        return prefs.index(player) < prefs.index(other)

    @abc.abstractmethod
    def _match(self, other):
        """A placeholder function for assigning the player to be matched to
        some other player."""

    @abc.abstractmethod
    def _unmatch(self, other):
        """A placeholder function for unassigning the player from its match
        with some other player."""

    @abc.abstractmethod
    def get_favourite(self):
        """A placeholder function for getting the player's favourite, feasible
        player."""

    @abc.abstractmethod
    def get_successors(self):
        """A placeholder function for getting the logically feasible
        'successors' of the player."""

    @abc.abstractmethod
    def check_if_match_is_unacceptable(self):
        """A placeholder for chacking the acceptability of the current
        match(es) of the player."""


class BaseGame(metaclass=abc.ABCMeta):
    """An abstract base class for facilitating various matching games.

    Parameters
    ----------
    clean
        Defaults to :code:`False`. If :code:`True`, when passing a set of
        players to create a game instance, they will be automatically cleaned.

    Attributes
    ----------
    matching
        After solving the game, a :code:`Matching` object is found here.
        Otherwise, :code:`None`.
    blocking_pairs
        After checking the stability of the game instance, a list of any pairs
        that block the stability of the matching is found here. Otherwise,
        :code:`None`.
    """

    def __init__(self, clean=False):

        self.matching = None
        self.blocking_pairs = None
        self.clean = clean

    def _remove_player(self, player, player_party, other_party):
        """Remove a player from the game instance as well as any relevant
        player preference lists."""

        party = vars(self)[player_party][:]
        party.remove(player)
        vars(self)[player_party].remove(player)
        for other in vars(self)[other_party]:
            if player in other.prefs:
                other._forget(player)

    def _check_inputs_player_prefs_unique(self, party):
        """Check that each player in :code:`party` has not ranked another
        player more than once. If so, and :code:`clean` is :code:`True`, then
        take the first instance they appear in the preference list."""

        for player in vars(self)[party]:
            unique_prefs = []
            for other in player.prefs:
                if other not in unique_prefs:
                    unique_prefs.append(other)
                else:
                    warnings.warn(
                        PreferencesChangedWarning(
                            f"{player} has ranked {other} multiple times."
                        )
                    )

            if self.clean:
                player.set_prefs(unique_prefs)

    def _check_inputs_player_prefs_all_in_party(self, party, other_party):
        """Check that each player in :code:`party` has ranked only players in
        :code:`other_party`. If :code:`clean`, then forget any extra
        preferences."""

        players = vars(self)[party]
        others = vars(self)[other_party]
        for player in players:

            for other in player.prefs:
                if other not in others:
                    warnings.warn(
                        PreferencesChangedWarning(
                            f"{player} has ranked a non-{other_party[:-1]}: "
                            f"{other}."
                        )
                    )
                    if self.clean:
                        player._forget(other)

    def _check_inputs_player_prefs_nonempty(self, party, other_party):
        """Make sure that each player in :code:`party` has a nonempty
        preference list of players in :code:`other_party`. If :code:`clean`,
        remove any such player."""

        for player in vars(self)[party]:

            if not player.prefs:
                warnings.warn(
                    PlayerExcludedWarning(
                        f"{player} has an empty preference list."
                    )
                )
                if self.clean:
                    self._remove_player(player, party, other_party)

    @abc.abstractmethod
    def solve(self):
        """ Placeholder for solving the given matching game. """

    @abc.abstractmethod
    def check_stability(self):
        """ Placeholder for checking the stability of the current galeshapley. """

    @abc.abstractmethod
    def check_validity(self):
        """ Placeholder for checking the validity of the current galeshapley. """


class BaseMatching(dict, metaclass=abc.ABCMeta):
    """An abstract base class for the storing and updating of a galeshapley.

    Attributes
    ----------
    dictionary : dict or None
        If not ``None``, a dictionary mapping a ``Player`` to one of: ``None``,
        a single ``Player`` or a list of ``Player`` instances.
    """

    def __init__(self, dictionary=None):

        self._data = {}
        if dictionary is not None:
            self._data.update(dictionary)

        super().__init__(self._data)

    def __repr__(self):

        return repr(self._data)

    def keys(self):

        return self._data.keys()

    def values(self):

        return self._data.values()

    def __getitem__(self, player):

        return self._data[player]

    @abc.abstractmethod
    def __setitem__(self, player, new_match):
        """ A placeholder function for how to update the galeshapley. """

    def _check_player_in_keys(self, player):
        """ Raise an error if :code:`player` is not in the dictionary. """

        if player not in self._data.keys():
            raise ValueError(f"{player} is not a key in this galeshapley.")

    def _check_new_valid_type(self, new, types):
        """Raise an error is :code:`new` is not an instance of one of
        :code:`types`."""

        if not isinstance(new, types):
            raise ValueError(f"{new} is not one of {types} and is not valid.")



from setuptools import find_packages, setup
import pathlib

here=pathlib.Path(__file__).parent.resolve()
long_description=(here / 'README.md').read_text(encoding='utf-8')

setup(
    name='galeshapley',
    version="0.1.4",
    description='Gale Shapley Algorithm',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/stamps-group/galeshapley',
    author='Gomes, J.M.',
    author_email='gomejm@ita.br',
    license='MIT',
    keywords=["game-theory gale-shapley matching-games"],
    install_requires=[],
    packages=find_packages("src"),
    package_dir={"": "src"},
    python_requires=">=3.5",
    setup_requires=['pytest-runner'],
    tests_require=["pytest", "hypothesis", "numpy"],
    test_suite='tests',
)



""" Functions for the SR algorithm. """
import warnings

from galeshapley.exceptions import NoStableMatchingWarning

from .util import _delete_pair


def first_phase(players):
    """Conduct the first phase of the algorithm where one-way proposals are
    made, and unpreferable pairs are forgotten."""

    free_players = players[:]
    while free_players:

        player = free_players.pop()
        favourite = player.get_favourite()

        current = favourite.matching
        if current is not None:
            favourite._unmatch()
            free_players.append(current)

        favourite._match(player)

        for successor in favourite.get_successors():
            _delete_pair(successor, favourite)
            if not successor.prefs and successor in free_players:
                free_players.remove(successor)

    return players


def locate_all_or_nothing_cycle(player):
    """Locate a cycle of (least-preferable, second-choice) pairs to be removed
    from the game."""

    lasts = [player]
    seconds = []
    while True:
        second_best = player.prefs[1]
        their_worst = second_best.prefs[-1]

        seconds.append(second_best)
        lasts.append(their_worst)

        player = their_worst

        if lasts.count(player) > 1:
            break

    idx = lasts.index(player)
    cycle = list(zip(lasts[idx + 1 :], seconds[idx:]))

    return cycle


def get_pairs_to_delete(cycle):
    """Based on an all-or-nothing cycle :math:`(x_1, y_1), \\ldots, (x_n, y_n)`,
    for each :math:`i = 1, \\ldots, n`, one must delete from the game all pairs
    :math:`(y_i, z)` such that :math:`y_i` prefers :math:`x_{i-1}` to :math:`z`
    where subscripts are taken modulo :math:`n`.

    This is an important point that is omitted from the original paper, but may
    be found in :cite:`GI89` (Section 4.2.3).

    The essential difference between this statement and that in :cite:`Irv85` is
    the removal of unpreferable pairs, identified using an all-or-nothing cycle,
    in addition to those contained in the cycle. Without doing so, tails of
    cycles can be removed rather than whole cycles, leaving some conflicting
    pairs in the game."""

    pairs = []
    for i, (_, right) in enumerate(cycle):

        left = cycle[(i - 1) % len(cycle)][0]
        successors = right.prefs[right.prefs.index(left) + 1 :]
        for successor in successors:
            pair = (right, successor)
            if pair not in pairs and pair[::-1] not in pairs:
                pairs.append((right, successor))

    return pairs


def second_phase(players):
    """Conduct the second phase of the algorithm where all-or-nothing cycles
    (rotations) are located and removed from the game."""

    player = next(p for p in players if len(p.prefs) > 1)
    while True:

        cycle = locate_all_or_nothing_cycle(player)
        pairs = get_pairs_to_delete(cycle)
        for player, other in pairs:
            _delete_pair(player, other)

        if any(p.prefs == [] for p in players):
            warnings.warn(
                NoStableMatchingWarning(
                    "The following players have emptied their preference list: "
                    f"{[p for p in players if not p.prefs]}"
                )
            )
            break

        try:
            player = next(p for p in players if len(p.prefs) > 1)
        except StopIteration:
            break

    for player in players:
        player._unmatch()
        if player.prefs:
            player._match(player.get_favourite())

    return players


def stable_duos(players):
    """Irving's algorithm :cite:`Irv85` that finds stable solutions to
    instances of SR if one exists. Otherwise, an incomplete matching is found.

    Parameters
    ----------
    players : list of Player
        The players in the game. Each must rank all other players.

    Returns
    -------
    matching : dict
        A dictionary of matches where the keys and values are given by the
        members of ``players``.
    """

    players = first_phase(players)

    if any(p.prefs == [] for p in players):
        warnings.warn(
            NoStableMatchingWarning(
                "The following players have been rejected by all others, "
                "emptying their preference list: "
                f"{[p for p in players if not p.prefs]}"
            )
        )

    if any(len(p.prefs) > 1 for p in players):
        players = second_phase(players)

    return {player: player.matching for player in players}



""" The HR game class and supporting functions. """
import copy
import warnings

from galeshapley import BaseGame, MultipleMatching
from galeshapley import Player as Part
from galeshapley.algorithms import base_container_part
from galeshapley.exceptions import (
    MatchingError,
    PlayerExcludedWarning,
    PreferencesChangedWarning,
)
from galeshapley.players import BaseContainer


class BaseContainerPart(BaseGame):
    """A class for solving instances of the base_container-part assignment
    problem (HR).

    In this case, a blocking pair is any part-base_container pair that satisfies
    **all** of the following:

        - They are present in each other's preference lists;
        - either the part is unmatched, or they prefer the base_container to their
          current match;
        - either the base_container is under-subscribed, or they prefer the part
          to at least one of their current matches.

    Parameters
    ----------
    parts : list of Player
        The parts in the matching game. Each part must rank a subset of
        those in :code:`base_containers`.
    base_containers : list of BaseContainer
        The base_containers in the matching game. Each base_container must rank all of (and
        only) the parts which rank it.
    clean : bool
        Indicator for whether the players of the game should be cleaned.
        Cleaning is reductive in nature, removing players from the game and/or
        other player's preferences if they do not meet the requirements of the
        game.

    Attributes
    ----------
    matching : Matching or None
        Once the game is solved, a matching is available as a :code:`Matching`
        object with the base_containers as keys and their part matches as values.
        Initialises as :code:`None`.
    blocking_pairs : list of (Player, BaseContainer) or None
        Initialises as `None`. Otherwise, a list of the part-base_container
        blocking pairs.
    """

    def __init__(self, parts, base_containers, clean=False):

        parts, base_containers = copy.deepcopy([parts, base_containers])
        self.parts = parts
        self.base_containers = base_containers
        self.clean = clean

        self._all_parts = parts
        self._all_base_containers = base_containers

        super().__init__(clean)
        self.check_inputs()

    @classmethod
    def create_from_dictionaries(
        cls, part_prefs, base_container_prefs, capacities, clean=False
    ):
        """Create an instance of :code:`BaseContainerPart` from two preference
        dictionaries and capacities. If :code:`clean=True` then remove players
        from the game and/or player preferences if they do not satisfy the
        conditions of the game."""

        parts, base_containers = _make_players(
            part_prefs, base_container_prefs, capacities
        )
        game = cls(parts, base_containers, clean)

        return game

    def solve(self, optimal="part"):
        """Solve the instance of HR using either the part- or
        base_container-oriented algorithm. Return the galeshapley."""

        self.matching = MultipleMatching(
            base_container_part(self.parts, self.base_containers, optimal)
        )
        return self.matching

    def check_validity(self):
        """ Check whether the current matching is valid. """

        unacceptable_issues = self._check_for_unacceptable_matches(
            "parts"
        ) + self._check_for_unacceptable_matches("base_containers")

        oversubscribed_issues = self._check_for_oversubscribed_players(
            "base_containers"
        )

        if unacceptable_issues or oversubscribed_issues:
            raise MatchingError(
                unacceptable_matches=unacceptable_issues,
                oversubscribed_base_containers=oversubscribed_issues,
            )

        return True

    def _check_for_unacceptable_matches(self, party):
        """Check that no player in `party` is matched to an unacceptable
        player."""

        issues = []
        for player in vars(self)[party]:
            issue = player.check_if_match_is_unacceptable(unmatched_okay=True)
            if isinstance(issue, list):
                issues.extend(issue)
            elif isinstance(issue, str):
                issues.append(issue)

        return issues

    def _check_for_oversubscribed_players(self, party):
        """ Check that no player in `party` is oversubscribed. """

        issues = []
        for player in vars(self)[party]:
            issue = player.check_if_oversubscribed()
            if issue:
                issues.append(issue)

        return issues

    def check_stability(self):
        """Check for the existence of any blocking pairs in the current
        matching, thus determining the stability of the galeshapley."""

        blocking_pairs = []
        for part in self.parts:
            for base_container in self.base_containers:
                if (
                    _check_mutual_preference(part, base_container)
                    and _check_part_unhappy(part, base_container)
                    and _check_base_container_unhappy(part, base_container)
                ):
                    blocking_pairs.append((part, base_container))

        self.blocking_pairs = blocking_pairs
        return not any(blocking_pairs)

    def check_inputs(self):
        """Give out warnings if any of the conditions of the game have been
        broken. If the :code:`clean` attribute is :code:`True`, then remove any
        such situations from the game."""

        self._check_inputs_player_prefs_unique("parts")
        self._check_inputs_player_prefs_unique("base_containers")

        self._check_inputs_player_prefs_all_in_party("parts", "base_containers")
        self._check_inputs_player_prefs_all_in_party("base_containers", "parts")

        self._check_inputs_player_prefs_all_reciprocated("base_containers")
        self._check_inputs_player_reciprocated_all_prefs(
            "base_containers", "parts"
        )

        self._check_inputs_player_prefs_nonempty("parts", "base_containers")
        self._check_inputs_player_prefs_nonempty("base_containers", "parts")

        self._check_inputs_player_capacity("base_containers", "parts")

    def _check_inputs_player_prefs_all_reciprocated(self, party):
        """Make sure that each player in :code:`party` has ranked only those
        players that have ranked it."""

        for player in vars(self)[party]:

            for other in player.prefs:
                if player not in other.prefs:
                    warnings.warn(
                        PreferencesChangedWarning(
                            f"{player} ranked {other} but they did not."
                        )
                    )
                    if self.clean:
                        player._forget(other)

    def _check_inputs_player_reciprocated_all_prefs(self, party, other_party):
        """Make sure that each player in :code:`party` has ranked all those
        players in :code:`other_party` that have ranked it."""

        players = vars(self)[party]
        others = vars(self)[other_party]
        for player in players:

            others_that_ranked = [
                other for other in others if player in other.prefs
            ]
            for other in others_that_ranked:
                if other not in player.prefs:
                    warnings.warn(
                        PreferencesChangedWarning(
                            f"{other} ranked {player} but they did not."
                        )
                    )
                    if self.clean:
                        other._forget(player)

    def _check_inputs_player_capacity(self, party, other_party):
        """Check that each player in :code:`party` has a capacity of at least
        one. If the :code:`clean` attribute is :code:`True`, remove any base_container
        that does not have such a capacity from the game."""

        for player in vars(self)[party]:
            if player.capacity < 1:
                warnings.warn(PlayerExcludedWarning(player))

                if self.clean:
                    self._remove_player(player, party, other_party)


def _check_mutual_preference(part, base_container):
    """ Determine whether two players each have a preference of the other. """

    return part in base_container.prefs and base_container in part.prefs


def _check_part_unhappy(part, base_container):
    """Determine whether a part is unhappy because they are unmatched, or
    they prefer the base_container to their current match."""

    return part.matching is None or part.prefers(
        base_container, part.matching
    )


def _check_base_container_unhappy(part, base_container):
    """Determine whether a base_container is unhappy because they are
    under-subscribed, or they prefer the part to at least one of their
    current matches."""

    return len(base_container.matching) < base_container.capacity or any(
        [base_container.prefers(part, match) for match in base_container.matching]
    )


def _make_players(part_prefs, base_container_prefs, capacities):
    """Make a set of parts and base_containers from the dictionaries given, and
    add their preferences."""

    part_dict, base_container_dict = _make_instances(
        part_prefs, base_container_prefs, capacities
    )

    for part_name, part in part_dict.items():
        prefs = [base_container_dict[name] for name in part_prefs[part_name]]
        part.set_prefs(prefs)

    for base_container_name, base_container in base_container_dict.items():
        prefs = [part_dict[name] for name in base_container_prefs[base_container_name]]
        base_container.set_prefs(prefs)

    parts = list(part_dict.values())
    base_containers = list(base_container_dict.values())

    return parts, base_containers


def _make_instances(part_prefs, base_container_prefs, capacities):
    """Create ``Player`` (part) and ``BaseContainer`` instances for the names in
    each dictionary."""

    part_dict, base_container_dict = {}, {}
    for part_name in part_prefs:
        part = Part(name=part_name)
        part_dict[part_name] = part
    for base_container_name in base_container_prefs:
        capacity = capacities[base_container_name]
        base_container = BaseContainer(name=base_container_name, capacity=capacity)
        base_container_dict[base_container_name] = base_container

    return part_dict, base_container_dict
